import sys

from rich.console import Console

from ..enums.terminal_out_file import TerminalOutFile

RICH_CONSOLE: dict[TerminalOutFile, Console] = {}


def get_rich_console(out_file: TerminalOutFile, width: int, force_terminal: bool) -> Console:
    """
    Get rich console, create if not exists

    :param out_file: terminal out file
    :param width: terminal width
    :param force_terminal: force terminal
    :return: rich console
    """

    global RICH_CONSOLE

    # check if rich console exists
    if out_file not in RICH_CONSOLE:
        # choose file
        if out_file == TerminalOutFile.STDOUT:
            file = sys.stdout
        elif out_file == TerminalOutFile.STDERR:
            file = sys.stderr
        else:
            file = sys.stderr

        # create rich console
        RICH_CONSOLE[out_file] = Console(file=file,
                                         force_terminal=force_terminal,
                                         width=width)

    rich_handler = RICH_CONSOLE[out_file]

    return rich_handler



from __future__ import annotations

import logging
from types import FunctionType


class BaseLogHandler(logging.Handler):
    def __new__(cls, *args, **kwargs):
        if hasattr(cls, '__post_init__'):
            # check if __post_init__ is abstract
            if getattr(cls.__post_init__, '__post_init__implemented__', False):
                # make method not abstract
                __init__ = cls.__init__

                def post_init_class_schema__init__(self, *a, **kw):
                    __init__(self, *a, **kw)
                    cls.__post_init__(self)

                cls.__init__ = post_init_class_schema__init__

        return super().__new__(cls)

    @classmethod
    def __init_subclass__(cls, **kwargs):
        __init__ = cls.__init__

        def post_init_class_schema__init__(self, *a, **kw):
            __init__(self, *a, **kw)
            cls.__post_init__(self)

        if not hasattr(cls, '__post_init__'):
            raise NotImplementedError(f"Class '{cls.__name__}' must implement '__post_init__' method")
        else:
            # check if __post_init__ is callable
            if not callable(cls.__post_init__):
                raise ValueError(f"__post_init__ for '{cls.__name__}' is not callable")

            # check if __post_init__ is __post_init__implemented__
            if getattr(cls.__post_init__, '__post_init__implemented__', False):
                return  # skip if abstract

            cls.__init__ = post_init_class_schema__init__

    def __init__(self, level=logging.NOTSET):
        super().__init__(level=level)
        self.header_width: int | None = None

    @staticmethod
    def implement_post_init():
        def decorator(func: FunctionType):
            func.__post_init__implemented__ = True
            return func

        return decorator

    @implement_post_init()
    def __post_init__(self):
        self.header_width = 110

        # wrap self.handle
        backup_format = self.format

        def format_wrapper(record):
            if record.msg == "[HEADER]" and hasattr(record, "header"):
                return self.format_header(record)
            else:
                return backup_format(record)

        self.format = format_wrapper


    def format_header(self, record) -> str:
        name: str = record.header["name"]
        desc: str = record.header["desc"]
        lines: list[str] = record.header["lines"]

        # prepare lines
        def prepare_line(l: str) -> str:
            if "\n" in l:
                raise ValueError("Line can't contain new line character")

            l_len = len(l) + 4
            if l_len > self.header_width:
                l_trimmed = l[: self.header_width - 4]
                l_other = l[self.header_width - 4:]
                l_trimmed_prepared = prepare_line(l_trimmed)
                l_other_prepared = prepare_line(l_other)
                return l_trimmed_prepared + l_other_prepared
            else:
                return f"║ {l}" + f"{' ' * (self.header_width - len(l) - 3)}" + "║\n"

        lines_prepared = []
        if lines is not None:
            for line in lines:
                for new_line in line.split("\n"):
                    lines_prepared.append(prepare_line(new_line))

        desc_prepared = ""
        if desc is not None:
            desc_prepared = "╠" + f"{'═' * (self.header_width - 2)}" + "╣\n"
            desc = f"description: {desc}"
            for new_line in desc.split("\n"):
                desc_prepared += prepare_line(new_line)
            desc_prepared += prepare_line(f"logger: {record.name.strip()}")

        # prepare header
        header_str = ""
        # time
        time_str = "═ " + self.formatter.formatTime(record) + " ═"
        time_added = False
        header_sum = self.header_width - 2
        line_sum = self.header_width - 2
        if len(header_str) + len(time_str) <= self.header_width:
            header_str += time_str
            header_sum -= len(time_str)
            time_added = True
        # level
        if time_added:
            level_str = " " + record.levelname + " ═"
        else:
            level_str = "═ " + record.levelname + " ═"
        if len(header_str) + len(level_str) <= self.header_width:
            header_str += level_str
            header_sum -= len(level_str)

        header = f"╔{header_str + '═' * header_sum}" + "╗\n"
        header += prepare_line(name)
        header += desc_prepared
        if len(lines_prepared) > 0:
            header += "╠" + f"{'═' * line_sum}" + "╣\n"
            header += "".join(lines_prepared)
        header += "╚" + f"{'═' * line_sum}" + "╝"

        # replace msg with header
        return header



from __future__ import annotations

import asyncio
import inspect
from typing import Callable, Type

_wrapper_self_of_cls = None


class Wrapper:
    # noinspection PyMethodParameters
    def __new__(_cls, *args, **kwargs):

        def _wrapper(self_of_cls=None, *a, **kw):
            global _wrapper_self_of_cls
            if not isinstance(self_of_cls, object):
                if not issubclass(self_of_cls, type):
                    _wrapper_self_of_cls = True
                else:
                    _wrapper_self_of_cls = self_of_cls
            else:
                _wrapper_self_of_cls = self_of_cls
            wrapper = _cls(*args, **kwargs)
            return wrapper(*a, **kw)

        global _wrapper_self_of_cls
        if _wrapper_self_of_cls is None:
            return _wrapper

        __init__ = _cls.__init__

        def __wrapper_init__(self, *a, **kw):
            global _wrapper_self_of_cls
            if _wrapper_self_of_cls is None:
                return  # Do nothing
            kw["_wrapper_self_of_cls"] = _wrapper_self_of_cls
            _wrapper_self_of_cls = None
            __init__(self, *a, **kw)

        __wrapper_init__.__wrapped_init__ = True

        if not getattr(__init__, "__wrapped_init__", False):
            _cls.__init__ = __wrapper_init__

        return super().__new__(_cls)

    def __init__(self, cls, method, hook=None, _wrapper_self_of_cls=None, **kwargs):
        self._cls = cls
        self._is_cls = True
        self._method = method
        self._hook = hook
        self._hook_disabled = False
        self._wrapper_self_of_cls = _wrapper_self_of_cls
        self._is_self = True
        self._kwargs = kwargs
        self._method_signature = inspect.signature(self.method)

        # check if cls in signature
        if "cls" not in self.method_signature.parameters.keys():
            self._is_cls = False
        # check if self in signature
        if "self" not in self.method_signature.parameters.keys():
            self._is_self = False

    def __str__(self):
        if self.cls is not None:
            return f"{self.cls.__name__}.{self.method.__name__} -> {self.kwargs}"
        return f"{self.method.__name__} -> {self.kwargs}"

    def __getattr__(self, item):
        if item.startswith("_"):
            return super().__getattribute__(item)
        if item in ["self", "cls", "method", "hook", "kwargs"]:
            return super().__getattribute__(item)
        return self._kwargs[item]

    def __setattr__(self, key, value):
        if key.startswith("_"):
            return super().__setattr__(key, value)
        if key in ["self", "cls", "method", "hook", "kwargs"]:
            return super().__setattr__(key, value)
        self._kwargs[key] = value

    def __getitem__(self, item):
        return self._kwargs[item]

    def __setitem__(self, key, value):
        self._kwargs[key] = value

    def __iter__(self):
        return iter(self._kwargs)

    def __call__(self, *args, **kwargs):
        is_hook_async = inspect.iscoroutinefunction(self.hook)
        is_method_async = inspect.iscoroutinefunction(self.method)

        if self.self is not None:
            if self.hook is not None and not self.hook_disabled:
                self._hook_disabled = True
                if not is_hook_async:
                    return self.hook(self, *args, **kwargs)
                return asyncio.get_running_loop().create_task(self.hook(self, *args, **kwargs))
            if not is_method_async:
                return self.method(self.self, *args, **kwargs)
            return asyncio.get_running_loop().create_task(self.method(self.self, *args, **kwargs))
        if self.cls is not None:
            if not is_method_async:
                return self.method(self.cls, *args, **kwargs)
            return asyncio.get_running_loop().create_task(self.method(self.cls, *args, **kwargs))
        if not is_method_async:
            return self.method(*args, **kwargs)
        return asyncio.get_running_loop().create_task(self.method(*args, **kwargs))

    @property
    def self(self):
        if not self._is_self:
            return None
        if isinstance(self._wrapper_self_of_cls, object):
            return self._wrapper_self_of_cls
        return None

    @self.setter
    def self(self, value):
        self._wrapper_self_of_cls = value

    @property
    def cls(self) -> Type | None:
        if not self._is_cls:
            return None
        return self._cls

    @cls.setter
    def cls(self, value):
        self._cls = value

    @property
    def method(self) -> Callable:
        return self._method

    @property
    def method_signature(self) -> inspect.Signature:
        return self._method_signature

    @property
    def hook(self) -> Callable:
        return self._hook

    @property
    def hook_disabled(self) -> bool:
        return self._hook_disabled

    @property
    def kwargs(self):
        return self._kwargs



from re import sub


def snake_case(s: str):
    """
    Convert a string to snake case.

    :param s: String to convert
    :return: String in snake case
    """

    return '_'.join(
        sub('([A-Z][a-z]+)', r' \1',
            sub('([A-Z]+)', r' \1',
                s.replace('-', ' '))).split()).lower()



from __future__ import annotations

import logging
import random
import string
from typing import Sequence, Type, Any

from controllogger.misc.easy_logger import easy_logger
from controllogger.misc.input_logger_config import InputLoggerConfig
from controllogger.misc.output_logger_config import OutputLoggerConfig
from controllogger.misc.logger_defaults_config import LoggerDefaultsConfig
from controllogger.misc.singleton import Singleton


class LoggerContext:
    def __init__(self,
                 context: str,
                 init: bool,
                 close: bool,
                 output_config: OutputLoggerConfig | dict[str, any] | Sequence[OutputLoggerConfig | dict[str, any]] = None,
                 input_config: InputLoggerConfig | dict[str, any] | Sequence[InputLoggerConfig | dict[str, any]] = None,
                 defaults_config: LoggerDefaultsConfig | dict[str, any] = None,
                 logger_cls: type[logging.Logger] = logging.Logger,
                 _parent_context: "LoggerContext" = None):
        # if context is None generate random context if not, check if context is in backups
        if context is None:
            generated_context = ""
            while self.control_logger.context_exists(generated_context) or generated_context == "":
                generated_context = "".join([random.choice(string.ascii_letters + string.digits) for _ in range(10)])
        self._context = context

        self._init = init
        self._close = close

        if defaults_config is not None:
            if type(defaults_config) is dict:
                defaults_config = LoggerDefaultsConfig(**defaults_config)
        self._logger_defaults_config = defaults_config

        self._parent_context = _parent_context
        self._child_contexts = []
        self._is_closed = False
        self._logger_cls = logger_cls
        self._context_exists = False

        # create output logger
        if output_config is not None:
            if type(output_config) is dict:
                output_config = OutputLoggerConfig(**output_config)
            if type(output_config) is Sequence:
                output_config = [OutputLoggerConfig(**c) if type(c) is dict else c for c in output_config]

            self.create_output_logger(config=output_config)

        # create input logger
        if input_config is not None:
            if type(input_config) is dict:
                input_config = InputLoggerConfig(**input_config)
            if type(input_config) is Sequence:
                input_config = [InputLoggerConfig(**c) if type(c) is dict else c for c in input_config]

            self.create_input_logger(config=input_config)

    def __repr__(self):
        return f"{self.__class__.__name__}(context={self.context}, init={self._init}, close={self._close})"

    def __str__(self):
        return f"{self.context}"

    def __call__(self, close: bool = False):
        if self.is_closed:
            raise ValueError("Reusing context with close=True is not allowed. Create your context with close=False and close it manually.")
        for context_logger in self.control_logger.context_logger:
            if not context_logger.open:
                raise ValueError(f"Logger '{context_logger.name}' already closed.")
        child_context = LoggerContext(context=self.context, init=False, close=close, _parent_context=self)
        self._child_contexts.append(child_context)
        return child_context

    def __enter__(self) -> "LoggerContext":
        self._context_exists = self.control_logger.context_exists(self.context)
        if self._context_exists:
            easy_logger.info(f"Skip entering context -> Already on context '{self.context}'")
            return self
        easy_logger.info(f"Entering context '{self.context}'")
        _active_contexts = getattr(self.control_logger, "_active_contexts")

        if self.control_logger.active_context is not None:
            # backup current context
            _active_contexts.insert(0, self.control_logger.active_context)

        # create new context handler
        setattr(self.control_logger, "_active_context", self)

        if self._init:
            # check if loggers with context already exist
            for context_logger in self.control_logger.context_logger:
                if self.context == context_logger.contexts:
                    raise ValueError(f"Context '{self.context}' already exists.")

        # attach all loggers with context
        self.attach()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._context_exists:
            easy_logger.info(f"Skip exiting context -> Context '{self.context}' opened by other context handler.")
            return
        if self._close:
            self.close()
        else:
            self.detach()

        # restore context handler
        if len(getattr(self.control_logger, "_active_contexts")) > 0:
            setattr(self.control_logger, "_active_context", getattr(self.control_logger, "_active_contexts").pop(0))
        else:
            setattr(self.control_logger, "_active_context", None)

    @property
    def context(self) -> str:
        return self._context

    @property
    def control_logger(self):
        """
        Returns the control logger singleton

        :return: ControlLogger
        """

        return Singleton.get_by_type("ControlLogger")

    @property
    def context_logger(self) -> list[logging.Logger | Any]:
        """
        Returns a list of all loggers with context

        :return: list of all loggers with context
        """

        return self.control_logger.context_logger

    @property
    def parent(self) -> "LoggerContext":
        return self._parent_context

    @property
    def children(self) -> list["LoggerContext"]:
        return self._child_contexts

    @property
    def is_closed(self) -> bool:
        if self.parent is None:
            return self._is_closed
        return self.parent.is_closed

    @is_closed.setter
    def is_closed(self, value: bool):
        if self.parent is None:
            self._is_closed = value
        else:
            self.parent.is_closed = value

    @property
    def logger_defaults_config(self) -> LoggerDefaultsConfig:
        """
        Returns the default logger config for the context

        :return: LoggerDefaultsConfig
        """

        output_logger_config_dict = self.control_logger.logger_defaults_config.dict(exclude_none=True)
        # find self in active contexts
        found = False
        for active_context in self.control_logger.active_contexts[::-1]:
            if not found:
                if active_context.context != self.context:
                    continue
                found = True
                continue
            output_logger_config_dict.update(active_context.logger_defaults_config.dict(exclude_none=True))

        if self._logger_defaults_config is not None:
            output_logger_config_dict.update(self._logger_defaults_config.dict(exclude_none=True))
        return LoggerDefaultsConfig(**output_logger_config_dict)

    @logger_defaults_config.setter
    def logger_defaults_config(self, value: LoggerDefaultsConfig | dict[str, any]):
        if type(value) is dict:
            value = LoggerDefaultsConfig(**value)
        self._logger_defaults_config = value

    @property
    def logger_cls(self) -> type[logging.Logger]:
        return self._logger_cls

    def create_output_logger(self,
                             config: OutputLoggerConfig | Sequence[OutputLoggerConfig] | dict | list[dict] = None,
                             logger_cls: Type[logging.Logger] = None) -> logging.Logger | list[logging.Logger]:
        """
        Creates output new logger for context

        :param config: OutputLoggerConfig or list of OutputLoggerConfig
        :param logger_cls: OutputLogger class to use
        :return: OutputLogger or list of OutputLogger
        """

        if logger_cls is None:
            logger_cls = self.logger_cls
        return self.control_logger.create_output_logger(config=config, logger_cls=logger_cls)

    def create_input_logger(self,
                            config: InputLoggerConfig | Sequence[InputLoggerConfig] | dict | list[dict] = None,
                            logger_cls: Type[logging.Logger] = None) -> logging.Logger | list[logging.Logger]:
        """
        Creates input new logger for context

        :param config: InputLoggerConfig or list of InputLoggerConfig
        :param logger_cls: InputLogger class to use
        :return:
        """

        if logger_cls is None:
            logger_cls = self.logger_cls
        return self.control_logger.create_input_logger(config=config, logger_cls=logger_cls)

    def attach(self) -> None:
        """
        Attach all loggers with context
        :return: None
        """
        for context_logger in self.context_logger:
            found = False
            for context in context_logger.contexts:
                if context.context == self.context:
                    found = True
                    break
            if not found:
                continue
            if context_logger.attached:
                continue
            context_logger.attach()

    def detach(self) -> None:
        """
        Detach all loggers with context
        :return: None
        """

        for context_logger in self.context_logger:
            found = False
            for context in context_logger.contexts:
                if context.context == self.context:
                    found = True
                    break
            if not found:
                continue
            if not context_logger.attached:
                continue
            context_logger.detach()

    def close(self):
        """
        Close all loggers with context
        :return: None
        """

        self.is_closed = True

        for context_logger in self.context_logger:
            found = False
            for context in context_logger.contexts:
                if context.context == self.context:
                    found = True
                    break
            if not found:
                continue
            if not context_logger.open:
                continue
            context_logger.__del__()



from __future__ import annotations

import logging
from datetime import datetime

from controllogger.enums.log_levels import LogLevels
from controllogger.enums.operator import Operator
from controllogger.handler.python_console_handler import PythonConsoleHandler
from controllogger.handler.tar_rotating_file_handler import TarRotatingFileHandler
from controllogger.logger.base import BaseLogger
from controllogger.misc.easy_logger import easy_logger
from controllogger.misc.output_logger_config import OutputLoggerConfig

# try to import rich console handler
try:
    from controllogger.handler.rich_console_handler import RichConsoleHandler
except Exception:
    RichConsoleHandler = None
    pass  # todo for now its okay but later we need to handle this


class OutputLogger(BaseLogger):
    _ConfigClass = OutputLoggerConfig
    _my_type = "Output Logger"
    _control_logger_type_field = "output_logger_type"

    def __init__(self, name: str = None, level: LogLevels | int = logging.NOTSET, config: _ConfigClass | dict[str, any] = None):
        self._use_control_logger_name = True
        super().__init__(name=name, level=level, config=config)

        # check if logger already exists
        if self.name in self.manager.loggerDict:
            raise ValueError(f"Logger '{self.name}' already exists.")

        # add logger to logging.root.manager.loggerDict
        self.manager.loggerDict[self.name] = self

        self._disable_filter = False

        # if console
        if self.config.console:
            # try to create rich console handler
            if RichConsoleHandler is not None:
                self.console_handler = RichConsoleHandler(console_outfile=self.config.console_outfile,
                                                          console_width=self.config.console_width,
                                                          console_force_terminal=self.config.console_force_terminal)
            else:
                self.console_handler = PythonConsoleHandler(console_outfile=self.config.console_outfile,
                                                            console_width=self.config.console_width)

            # set console_header_width to config.console_width if console_header_width is greater than config.console_width
            console_header_width = self.config.header_width
            if console_header_width > self.config.console_width:
                console_header_width = self.config.console_width

            # set console header width
            self.console_handler.header_width = console_header_width

            # set console handler level to NOTSET to allow all records to be passed to console handler
            self.console_handler.setLevel(logging.NOTSET)

            # create console handler formatter
            self.console_formatter = logging.Formatter(self.config.console_format)

            # set console handler formatter
            self.console_handler.setFormatter(self.console_formatter)

            # add console handler to self
            self.addHandler(self.console_handler)

        # if file
        if self.config.file:
            # create directory if not exists
            self.config.file_path.parent.mkdir(parents=True, exist_ok=True)

            # create file handler
            self.file_handler = TarRotatingFileHandler(
                filename=self.config.file_path,
                backup_count=self.config.file_backup_count,
                max_bytes=self.config.file_max_size,
                archive_backup_count=self.config.file_archive_backup_count,
                encoding=self.config.file_encoding,
                delay=self.config.file_delay,
                mode=self.config.file_mode,
            )

            # set file_header_width
            self.file_handler.header_width = self.config.header_width

            # set file handler level to NOTSET to allow all records to be passed to file handler
            self.file_handler.setLevel(logging.NOTSET)

            # create file handler formatter
            self.file_formatter = logging.Formatter(self.config.file_format)

            # set file handler formatter
            self.file_handler.setFormatter(self.file_formatter)

            # add file handler to self
            self.addHandler(self.file_handler)

        # if db
        if self.config.db:
            # create db handler
            self.db_handler = DbLogHandler(log_document=self.config.db_document,
                                           buffer_max_lines=self.config.db_buffer_max_lines,
                                           buffer_max_time=self.config.db_buffer_max_time)

            # set db_header_width
            self.db_handler.header_width = self.config.header_width

            # set db handler level to NOTSET to allow all records to be passed to db handler
            self.db_handler.setLevel(logging.NOTSET)

            # create db handler formatter
            self.db_formatter = logging.Formatter(self.config.db_format)

            # set db handler formatter
            self.db_handler.setFormatter(self.db_formatter)

            # add db handler to self
            self.addHandler(self.db_handler)

        # add logger to control logger
        self.attach()

    def __del__(self):
        # remove logger from control logger
        if not self.open:
            return
        super().__del__()
        self.detach()
        # delete logger from logging.root.manager.loggerDict
        del self.control_logger.all_logger[self.name]

    def __str__(self) -> str:
        r = f"{self.__class__.__name__}(name={self.name}, level={LogLevels(self.level).name}, "
        if self.closed:
            r = r[:-2]
            r += ") CLOSED"
            return r
        if not self.attached:
            r = r[:-2]
            r += ") DETACHED"
            return r
        if len(self.contexts) > 0:
            r += f"contexts=({', '.join([str(context) for context in self.contexts])}), "
        if self.config.console:
            console = self.config.console_outfile.value
            r += f"console --> {console}, "

        if self.config.file:
            file_path = str(self.config.file_path)
            r += f"file --> {file_path}, "
        if self.config.db:
            db_document = "self.db_document.__class__.__name__"
            r += f"db --> {db_document}, "
        r = r[:-2]
        r += ")"

        return r

    def print_logger_initialized(self):
        if self.config.log_init:
            self._disable_filter = True
            self.print_header(
                name=f"{self._my_type} '{self.name}' initialized at {datetime.now()}",
                level=LogLevels.DEBUG,
                lines=[str(self)],
            )
            self._disable_filter = False

    def attach(self) -> None:
        """
        Attach the output logger to the control logger
        :return: None
        """

        if self.attached:
            raise ValueError(f"Logger '{self.name}' already attached.")
        if self.name in self.control_logger.output_loggers.keys():
            raise ValueError(f"Output logger for '{self.name}' already exists.")
        self.control_logger.output_loggers[self.name] = self
        super().attach()

    def detach(self) -> None:
        """
        Detach the output logger from the control logger
        :return: None
        """

        if not self.attached:
            raise ValueError(f"Logger '{self.name}' already detached.")
        if self.name not in self.control_logger.output_loggers.keys():
            raise ValueError(f"Output logger for '{self.name}' not found.")
        super().detach()
        del self.control_logger.output_loggers[self.name]

    @property
    def attached(self) -> bool:
        """
        Checks if output logger is attached to control logger

        :return: True if attached, False if not
        """

        for logger_name in self.control_logger.output_loggers.keys():
            if logger_name == self.name:
                return True
        return False

    def handle(self, record):
        # check if last resort enabled and set level to 999
        backup_last_resort_level = logging.lastResort.level
        if not self.config.last_resort:
            logging.lastResort.level = 999

        if (not self.disabled) and self.filter(record):
            self.callHandlers(record)

        # reset last resort level
        if not self.config.last_resort:
            logging.lastResort.level = backup_last_resort_level

    def filter(self, record) -> bool:
        if self._disable_filter:
            return True
        if self.config.filter:
            return self._filter(record)
        else:
            return True

    def _filter(self, record) -> bool:
        out = False
        for log_filter in self.config.filter:
            try:
                record_value = getattr(record, log_filter.key)

                # ToDo: at the moment this simple implementation is enough but but we have plans to conquer the whole world with this filter so keep gespannt wie n Flitzebogen

                # EQUALS
                if log_filter.operator == Operator.EQUALS:
                    if log_filter.value == record_value:
                        out = True
                        break
                # NOT_EQUALS
                elif log_filter.operator == Operator.NOT_EQUALS:
                    if log_filter.value != record_value:
                        out = True
                        break
                # CONTAINS
                elif log_filter.operator == Operator.CONTAINS:
                    if log_filter.value in record_value:
                        out = True
                        break
                # NOT_CONTAINS
                elif log_filter.operator == Operator.NOT_CONTAINS:
                    if log_filter.value not in record_value:
                        out = True
                        break
                # STARTS_WITH
                elif log_filter.operator == Operator.STARTS_WITH:
                    if record_value.startswith(log_filter.value):
                        out = True
                        break
                # ENDS_WITH
                elif log_filter.operator == Operator.ENDS_WITH:
                    if record_value.endswith(log_filter.value):
                        out = True
                        break
                # GREATER_THAN
                elif log_filter.operator == Operator.GREATER_THAN:
                    if log_filter.value > record_value:
                        out = True
                        break
                # GREATER_THAN_OR_EQUAL
                elif log_filter.operator == Operator.GREATER_THAN_OR_EQUAL:
                    if log_filter.value >= record_value:
                        out = True
                        break
                # LESS_THAN
                elif log_filter.operator == Operator.LESS_THAN:
                    if log_filter.value < record_value:
                        out = True
                        break
                # LESS_THAN_OR_EQUAL
                elif log_filter.operator == Operator.LESS_THAN_OR_EQUAL:
                    if log_filter.value <= record_value:
                        out = True
                        break
                else:
                    raise ValueError(f"Operator '{log_filter.operator}' not supported")

            except Exception as e:
                easy_logger.exception(f"Error while filtering log record: {e}")
                out = True

        return out



import logging
import sys

from controllogger.enums.terminal_out_file import TerminalOutFile
from controllogger.handler.base import BaseLogHandler


class PythonConsoleHandler(logging.StreamHandler, BaseLogHandler):
    def __init__(self, console_outfile: TerminalOutFile, console_width: int):
        if TerminalOutFile.STDOUT == console_outfile:
            stream = sys.stdout
        elif TerminalOutFile.STDERR == console_outfile:
            stream = sys.stderr
        else:
            raise ValueError(f"Unknown console_outfile: {console_outfile}")

        super().__init__(stream=stream)

        self.console_outfile = console_outfile
        self.console_width = console_width

    def handle(self, record):
        # check if any msg line is longer than console_width
        reorganize_msg = False
        for line in record.msg.splitlines():
            if len(line) > self.console_width:
                reorganize_msg = True
                break

        # reorganize msg
        if reorganize_msg:
            new_msg_lines = []
            for line in record.msg.splitlines():
                if len(line) > self.console_width:
                    for i in range(0, len(line), self.console_width):
                        new_msg_lines.append(line[i:i + self.console_width])
                else:
                    new_msg_lines.append(line)
            new_msg = "\n".join(new_msg_lines)
            record.msg = new_msg
        logging.StreamHandler.handle(self, record)



import logging

from controllogger.logger.base_easy import BaseEasyLogger


class EasyLogger(BaseEasyLogger):
    def __init__(self):
        super().__init__(name="easylogger", level=logging.WARNING)

        self.easy_logger_handler = logging.StreamHandler()
        self.easy_logger_handler_formatter = logging.Formatter(
            '[%(asctime)s][%(name)s][%(levelname)s] %(srcRelativePathname)s(%(srcLineno)s) -> %(relativePathname)s(%(lineno)s) - %(message)s')
        self.easy_logger_handler.setFormatter(self.easy_logger_handler_formatter)
        self.addHandler(self.easy_logger_handler)


easy_logger = EasyLogger()



from __future__ import annotations

import inspect
import json
import logging
import time
from json import JSONEncoder
from typing import Type, Callable, Sequence, Any

from controllogger.enums.log_levels import LogLevels
from controllogger.logger.input import InputLogger
from controllogger.logger.output import OutputLogger
from controllogger.misc.context import LoggerContext
from controllogger.misc.easy_logger import easy_logger
from controllogger.misc.easy_logger_config import EasyLoggerConfig
from controllogger.misc.input_logger_config import InputLoggerConfig
from controllogger.misc.last_resort import last_resort
from controllogger.misc.log_function_config import LogFuntionConfig
from controllogger.misc.output_logger_config import OutputLoggerConfig
from controllogger.misc.logger_defaults_config import LoggerDefaultsConfig
from controllogger.misc.singleton import Singleton
from controllogger.misc.snake_case import snake_case

try:
    from rich.pretty import pretty_repr
except ImportError:
    def pretty_repr(obj, indent_size=2) -> str:
        class LogJSONEncoder(JSONEncoder):
            def default(self, o):
                return repr(o)

        if isinstance(obj, dict):
            pretty_obj_str = json.dumps(obj, indent=indent_size, cls=LogJSONEncoder)
        elif hasattr(obj, "__dict__"):
            pretty_obj_str = json.dumps({k: getattr(obj, k) for k in obj.__slots__}, indent=indent_size, cls=LogJSONEncoder)
        else:
            raise TypeError(f"Object '{obj}' is not pretty repr-able.")
        return pretty_obj_str


    easy_logger.warning("Could not import rich.pretty.pretty_repr. Using default pretty_repr.")


class ControlLogger(logging.Logger, metaclass=Singleton):
    """
    Control logger for easy logger.

    This logger is used to control all other loggers.
    You can create input and output loggers with this logger.
    It also handles the context for input and output loggers.
    """

    def __init__(self, easy_logger_config: EasyLoggerConfig | dict[str, any] = None):
        # convert dict to EasyLoggerConfig
        if easy_logger_config is None:
            easy_logger_config = EasyLoggerConfig()
        elif type(easy_logger_config) is dict:
            easy_logger_config = EasyLoggerConfig(**easy_logger_config)

        self.config = easy_logger_config

        super().__init__(self.config.name, self.config.level)
        self.propagate = True
        self.disabled = False

        # setup last_resort logger
        if self.config.last_resort:
            last_resort.setLevel(logging.DEBUG)
        else:
            last_resort.setLevel(logging.ERROR)

        # log first message
        easy_logger.info(f"Initializing {self.__class__.__name__} with config:")
        easy_logger.debug(self.config.json(indent=4))

        self._input_loggers = {}
        self._output_loggers = {}

        # context variables
        self._active_context: LoggerContext | None = None
        self._active_contexts: list[LoggerContext] = []
        self._decorator_contexts: list[LoggerContext] = []

        # set logger classes
        self._temp_output_logger_cls: list[type[OutputLogger]] = []
        self._temp_input_logger_cls: list[type[InputLogger]] = []

        # current class_logger_cls
        self._class_logger_cls: type[Any] | None = None

        # create output loggers
        self.create_output_logger(self.config.output_loggers)

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    def __str__(self):
        _str = f"{self.__class__.__name__}("
        _str += f"name={self.name}, "
        _str += f"level={LogLevels(self.level).name}, "
        if not self.propagate:
            _str += f"propagate={self.propagate}, "
        if self.disabled:
            _str += f"disabled={self.disabled}, "
        if len(self.active_contexts) > 0:
            _str += f"active_contexts=[{', '.join([str(context) for context in self.active_contexts])}], "

        _str = _str[:-2]
        _str += ")"

        return _str

    @property
    def all_logger(self) -> dict[str, logging.Logger]:
        """
        Returns all loggers from logging.root.manager.loggerDict
        Note: no copy is returned, so be careful with this function.

        :return:
        """

        return logging.root.manager.loggerDict

    def logger_exists(self, name: str) -> bool:
        """
        Checks if logger exists.

        :param name: Name of the logger
        :return: True if logger exists, False if not
        """

        logger = self.all_logger.get(name, None)
        if logger is None:
            return False
        if type(logger) is logging.PlaceHolder:
            return False
        return True

    def handle(self, record) -> None:
        """
        Handles log records to output loggers by context.
        :param record: Current log record
        :return: None
        """

        contexts = getattr(record, "contexts")
        record_handled = False
        for context_logger in self.context_logger:
            # skip all InputLogger
            if isinstance(context_logger, InputLogger):
                continue

            # check if context_logger has context
            if len(context_logger.contexts) > 0:
                # check if context_logger has the right context
                found = False
                for context in contexts:
                    if context in context_logger.contexts:
                        found = True
                        break
                if not found:
                    continue  # skip if context not found
            else:  # context_logger has no context so it handles all records
                pass

            # check if context_logger is attached
            if context_logger.attached:
                record_handled = True
                context_logger.handle(record)

        # if record not handled, log to last resort
        if not record_handled:
            easy_logger.debug(f"ControlLogger: Could not handle record '{record}' because no context logger available.")
            last_resort.handle(record)

    @property
    def logger_defaults_config(self) -> LoggerDefaultsConfig:
        """
        Returns the default logger config.

        :return: LoggerDefaultsConfig
        """

        config_dict = self.config.dict(exclude_none=True)

        # remove output_loggers
        del config_dict["output_loggers"]

        # remove input_logger_cls
        del config_dict["input_logger_cls"]

        # remove output_logger_cls
        del config_dict["output_logger_cls"]

        return LoggerDefaultsConfig(**config_dict)

    @classmethod
    def get_logger(cls, name: str) -> InputLogger:
        """
        A Pythonic way to get a logger. And for better type hinting to InputLogger.
        :param name:
        :return: InputLogger
        """

        get_logger = getattr(logging, "getLogger")
        return get_logger(name)

    @property
    def input_loggers(self) -> dict[str, InputLogger]:
        """
        Returns all input loggers.
        :return: list of input loggers
        """

        return self._input_loggers

    def set_temp_input_logger_cls(self, logger_cls: Type[InputLogger] | Sequence[Type[InputLogger]] = None) -> None:
        """
        Sets the temporary input logger class.
        :param logger_cls: InputLogger class
        :return: None
        """

        if logger_cls is not None:
            if not isinstance(logger_cls, Sequence):
                logger_cls = [logger_cls]
        else:
            logger_cls = []

        self._temp_input_logger_cls = logger_cls

    def reset_temp_input_logger_cls(self) -> None:
        """
        Resets the temporary input logger class.
        :return: None
        """

        self._temp_input_logger_cls = []

    @property
    def input_logger_type(self) -> type:
        """
        Returns the input logger type.

        :return: InputLogger type
        """

        logger_classes = []
        for input_logger_cls in self._temp_input_logger_cls:
            logger_classes.append(input_logger_cls)
        if self.config.output_logger_cls is not None:
            logger_classes.append(self.config.input_logger_cls)
        input_logger_type = type(InputLogger.__name__, tuple(logger_classes), {})
        return input_logger_type

    def create_input_logger(self,
                            config: InputLoggerConfig | Sequence[InputLoggerConfig] | dict | list[dict] = None,
                            logger_cls: Type[InputLogger] | Sequence[Type[InputLogger]] = None) -> InputLogger | list[InputLogger]:
        """
        Creates input new logger.

        :param config: InputLoggerConfig or list of OutputLoggerConfig
        :param logger_cls: InputLogger class to use
        :return: InputLogger or list of InputLogger
        """

        if not isinstance(config, Sequence):
            config = [config]

        self.set_temp_input_logger_cls(logger_cls=logger_cls)
        input_logger = []
        for c in config:
            input_logger.append(InputLogger(config=c))
        self.reset_temp_input_logger_cls()

        if len(input_logger) == 1:
            return input_logger[0]
        else:
            return input_logger

    @property
    def output_loggers(self) -> dict[str, OutputLogger]:
        """
        Returns all output loggers.

        :return: list of output loggers
        """

        return self._output_loggers

    def set_temp_output_logger_cls(self, logger_cls: Type[OutputLogger] | Sequence[Type[OutputLogger]] = None) -> None:
        """
        Sets the temporary output logger class.
        :param logger_cls: OutputLogger class
        :return: None
        """

        if logger_cls is not None:
            if not isinstance(logger_cls, Sequence):
                logger_cls = [logger_cls]
        else:
            logger_cls = []

        self._temp_output_logger_cls = logger_cls

    def reset_temp_output_logger_cls(self) -> None:
        """
        Resets the temporary output logger class.
        :return: None
        """

        self._temp_output_logger_cls = []

    @property
    def output_logger_type(self) -> type:
        """
        Returns the output logger type.
        :return: OutputLogger type
        """

        logger_classes = []
        for output_logger_cls in self._temp_output_logger_cls:
            logger_classes.append(output_logger_cls)
        if self.config.output_logger_cls is not None:
            logger_classes.append(self.config.output_logger_cls)
        output_logger_type = type(OutputLogger.__name__, tuple(logger_classes), {})
        return output_logger_type

    def create_output_logger(self,
                             config: OutputLoggerConfig | Sequence[OutputLoggerConfig] | dict | list[dict] = None,
                             logger_cls: Type[OutputLogger] | Sequence[Type[OutputLogger]] = None) -> OutputLogger | list[OutputLogger]:
        """
        Creates output new logger.

        :param config: OutputLoggerConfig or list of OutputLoggerConfig
        :param logger_cls: OutputLogger class to use
        :return: OutputLogger or list of OutputLogger
        """

        if not isinstance(config, Sequence):
            config = [config]

        self.set_temp_output_logger_cls(logger_cls=logger_cls)
        output_logger = []
        for c in config:
            output_logger.append(OutputLogger(config=c))
        self.reset_temp_output_logger_cls()

        if len(output_logger) == 1:
            return output_logger[0]
        else:
            return output_logger

    @property
    def active_contexts(self) -> list[LoggerContext]:
        """
        Returns all active contexts.
        Note: a copy is returned, so it's safe to use the output.

        :return: list of active contexts
        """

        context = self._active_contexts.copy()
        if self._active_context is not None:
            context.append(self._active_context)
        return context

    @property
    def active_context(self) -> LoggerContext:
        """
        Returns the active context.

        :return: Current active context
        """

        return self._active_context

    @property
    def decorator_contexts(self) -> list[LoggerContext]:
        """
        Returns all decorator contexts.
        Note: a copy is returned, so it's safe to use the output.

        :return: list of decorator contexts
        """

        return self._decorator_contexts.copy()

    @property
    def context_logger(self) -> list[InputLogger | OutputLogger]:
        out = []
        for logger_name in logging.root.manager.loggerDict.keys():
            logger = logging.root.manager.loggerDict[logger_name]
            if isinstance(logger, InputLogger) or isinstance(logger, OutputLogger):
                is_context_logger = False
                if len(logger.contexts) == 0:
                    is_context_logger = True
                for context in logger.contexts:
                    found = False
                    for active_context in self.active_contexts:
                        if context.context == active_context.context:
                            found = True
                            break
                    if found:
                        is_context_logger = True
                        break
                if is_context_logger:
                    out.append(logger)
        return out

    def context_exists(self, context: str) -> bool:
        for context_handler in self.active_contexts:
            if context_handler.context == context:
                return True
        return False

    def get_function_context(self, context: str) -> LoggerContext:
        function_context = None
        for function_context in self.decorator_contexts:
            if function_context.context == context:
                break
        if function_context is None:
            raise ValueError(f"Could not find function context '{context}'.")
        return function_context

    def __call__(self,
                 context: str = None,
                 init: bool = True,
                 close: bool = True,
                 output_config: OutputLoggerConfig | dict[str, any] | Sequence[OutputLoggerConfig | dict[str, any]] = None,
                 input_config: InputLoggerConfig | dict[str, any] | Sequence[InputLoggerConfig | dict[str, any]] = None,
                 defaults_config: LoggerDefaultsConfig | dict[str, any] = None,
                 logger_cls: Type[logging.Logger] = None) -> LoggerContext:
        """
        Creates a new context.

        :param context: Context name
        :param init: Initialize context
        :param close: Close context after use
        :param output_config: OutputLoggerConfig or list of OutputLoggerConfig to create new output loggers with context
        :param input_config: InputLoggerConfig or list of InputLoggerConfig to create new input loggers with context
        :param defaults_config: LoggerDefaultsConfig to create new loggers in context
        :param logger_cls: OutputLogger class to use
        :return: LoggerContext
        """

        return LoggerContext(context=context,
                             init=init,
                             close=close,
                             output_config=output_config,
                             input_config=input_config,
                             defaults_config=defaults_config,
                             logger_cls=logger_cls)

    @staticmethod
    def _get_var_from_signature(signature: inspect.Signature, cls_type: Type | None, prefer_name: list[str], _matched: list[str]) -> tuple[list[str], str]:
        # fist looking fo type annotation
        if cls_type is not None:
            for param in signature.parameters.values():
                # skip all marked parameters
                if param.name in _matched:
                    continue
                if issubclass(param.annotation, cls_type):
                    # mark parameter as matched
                    _matched.append(param.name)
                    return _matched, param.name

        # if not found, looking for name
        for param in signature.parameters.values():
            # skip all marked parameters
            if param.name in _matched:
                continue
            # skip if a type annotation is set
            if param.annotation is not inspect.Parameter.empty:
                continue
            if param.name in prefer_name:
                # mark parameter as matched
                _matched.append(param.name)
                return _matched, param.name

        # if not found, get first parameter
        for param in signature.parameters.values():
            # skip all marked parameters
            if param.name in _matched:
                continue
            # skip if a type annotation is set
            if param.annotation is not inspect.Parameter.empty:
                continue
            # mark parameter as matched
            _matched.append(param.name)
            return _matched, param.name

        # if not found, raise error
        if cls_type is None:
            raise ValueError(f"Could not find parameter {prefer_name} in signature '{signature}'")
        raise ValueError(f"Could not find parameter of type '{cls_type.__name__}' in signature '{signature}'")

    @staticmethod
    def _parse_dict(*args, callback_signature: inspect.Signature, **kwargs) -> dict:

        parsed_kwargs = {}
        arg_index = 0

        def arg_value(ai):
            if arg_index >= len(args):
                a = "..."
            else:
                a = args[arg_index]
            ai += 1
            return a, ai

        def kwarg_value(p):
            if param.name in kwargs:
                return kwargs[param.name]
            if param.default is not inspect.Parameter.empty:
                return param.default
            return "..."

        for param in callback_signature.parameters.values():
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                parsed_kwargs[param.name] = args[arg_index:]
                arg_index = len(args)
            elif param.kind == inspect.Parameter.VAR_KEYWORD:
                parsed_kwargs[param.name] = kwargs
            elif param.kind == inspect.Parameter.POSITIONAL_ONLY:
                parsed_kwargs[param.name], arg_index = arg_value(arg_index)
            elif param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                if param.name not in kwargs:
                    parsed_kwargs[param.name], arg_index = arg_value(arg_index)
                else:
                    parsed_kwargs[param.name] = kwarg_value(param)
            elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                kwarg_value(param)

        return parsed_kwargs

    def _log_function(self,
                      *args,
                      func_signature: inspect.Signature,
                      func_or_cls: Callable | Type,
                      self_or_cls: str | None,
                      input_logger: InputLogger,
                      func_config: LogFuntionConfig,
                      start: float = None,
                      **kwargs) -> Any:
        # parse dict
        parse_dict = self._parse_dict(callback_signature=func_signature,
                                      *args,
                                      **kwargs)

        # get func_name name
        func_name = func_or_cls.__name__
        if "self" == self_or_cls:
            func_name = f"{parse_dict['self'].__class__.__name__}.{func_or_cls.__name__}"
        elif "cls" == self_or_cls:
            func_name = f"{parse_dict['cls'].__name__}.{func_or_cls.__name__}"
        elif self_or_cls is None:
            func_name = func_or_cls.__name__
        func_name += "()"

        # parse start message
        if func_config.start_msg is not None:
            s_msg = func_config.start_msg.format(**parse_dict)
        else:
            s_msg = ""

        # log header
        if func_config.header:
            pretty_parse_dict = pretty_repr(parse_dict, indent_size=2).split("\n")
            input_logger.print_header(
                name=f"{func_name}",
                desc=func_or_cls.__doc__,
                lines=[f"Arguments:", *pretty_parse_dict],
                level=func_config.header_footer_level,
            )

        # log start message
        if s_msg:
            input_logger.log(func_config.start_end_msg_level, s_msg)

        # call func
        result = func_or_cls(*args, **kwargs)

        # perform time measurement
        end = None
        measure_time_str = ""
        if func_config.measure_time:
            end = time.perf_counter()
            measure_time_str = f"Time elapsed: {end - start:.2f} seconds"

        # parse end message
        if func_config.end_msg is not None:
            e_msg = func_config.end_msg.format(**parse_dict)
        else:
            e_msg = ""

        # log end message
        if e_msg:
            input_logger.log(func_config.start_end_msg_level, e_msg)

        # log footer
        if func_config.footer:
            input_logger.print_header(
                name=f"{func_name} called",
                desc=func_or_cls.__doc__,
                lines=[f"Result: {result}", measure_time_str],
                level=func_config.header_footer_level,
            )
        else:
            if func_config.measure_time:
                input_logger.log(func_config.header_footer_level, measure_time_str)

        return result

    def _get_sig_vars(self,
                      func_signature: inspect.Signature,
                      pass_logger_context: str | None = ...,
                      pass_input_logger: str | None = ...,
                      _matched: list[str] = None) -> tuple[list[str], str, str]:
        # init matched
        if _matched is None:
            _matched = []

        # get input logger var from signature
        prefer_input_logger_name = ["l", "lg", "logger"]
        if pass_input_logger is not None:
            prefer_input_logger_name = [pass_input_logger]
        if pass_input_logger is Ellipsis:
            try:
                _matched, _pass_input_logger = self._get_var_from_signature(signature=func_signature,
                                                                            cls_type=InputLogger,
                                                                            prefer_name=prefer_input_logger_name,
                                                                            _matched=_matched)
            except ValueError:
                _pass_input_logger = None
        else:
            _pass_input_logger = pass_input_logger

        # get contex var from signature
        prefer_logger_context_name = ["c", "ctx", "context", "logger_context"]
        if pass_logger_context is not None:
            prefer_logger_context_name = [pass_logger_context]
        if pass_logger_context is Ellipsis:
            try:
                _matched, _pass_logger_context = self._get_var_from_signature(signature=func_signature,
                                                                              cls_type=LoggerContext,
                                                                              prefer_name=prefer_logger_context_name,
                                                                              _matched=_matched)
            except ValueError:
                _pass_logger_context = None
        else:
            _pass_logger_context = pass_logger_context

        return _matched, _pass_input_logger, _pass_logger_context

    def function_logger(self,
                        func_config: LogFuntionConfig | dict[str, any] = None,
                        name: str = None,
                        context: str = None,
                        output_config: OutputLoggerConfig | dict[str, any] | Sequence[OutputLoggerConfig | dict[str, any]] = None,
                        input_config: InputLoggerConfig | dict[str, any] | Sequence[InputLoggerConfig | dict[str, any]] = None,
                        defaults_config: LoggerDefaultsConfig | dict[str, any] = None,
                        pass_logger_context: str | None = ...,
                        pass_input_logger: str | None = ...) -> Callable:

        if func_config is None:
            func_config = LogFuntionConfig()
        elif type(func_config) is dict:
            func_config = LogFuntionConfig(**func_config)
        elif not isinstance(func_config, LogFuntionConfig):
            raise TypeError(f"func_config must be of type {LogFuntionConfig.__name__} or dict.")

        def decorator(func: Callable | Type) -> Callable:
            # check if function or type
            if inspect.isclass(func):
                raise NotImplementedError(f"Cant decorate class with {self.__class__.__name__}.function_logger(). Please use {self.__class__.__name__}.class_logger() instead.")

            # init matched
            _matched = []

            if isinstance(func, classmethod):
                raise RuntimeError(f"Please decorate first with @{self.__class__.__name__}.function_logger() then with @classmethod.")
            if isinstance(func, staticmethod):
                raise RuntimeError(f"Please decorate first with @{self.__class__.__name__}.function_logger() then with @staticmethod.")
            if isinstance(func, property):
                raise RuntimeError(f"Can't decorate property with {self.__class__.__name__}.function_logger(). Please use {self.__class__.__name__}.class_logger_method() instead.")

            # get signature
            func_signature = inspect.signature(func)

            # check if self or cls is in signature
            try:
                _matched, _self_or_cls = self._get_var_from_signature(signature=func_signature,
                                                                      cls_type=None,
                                                                      prefer_name=["self", "cls"],
                                                                      _matched=_matched)
            except ValueError:
                _self_or_cls = None

            if _self_or_cls == "self":
                if func.__name__ == "__init__":
                    raise NotImplementedError(
                        f"Cant decorate class method __init__ with {self.__class__.__name__}.function_logger(). Please use {self.__class__.__name__}.class_logger() instead.")
                raise NotImplementedError(
                    f"Cant decorate class method with {self.__class__.__name__}.function_logger(). Please use {self.__class__.__name__}.class_logger_method() instead.")
            elif _self_or_cls == "cls":
                ...
            else:
                _self_or_cls = None
                _matched = []

            # set default values
            _name = name
            if name is None:
                _name = snake_case(func.__name__)

            _context = context
            if context is None:
                _context = func.__name__

            # check if function or type
            easy_logger.debug(f"Creating context '{_context}' with name '{_name}' for function '{func.__name__}'")

            _pass_input_logger = pass_input_logger
            _pass_logger_context = pass_logger_context
            _matched, _pass_input_logger, _pass_logger_context = self._get_sig_vars(func_signature=func_signature,
                                                                                    pass_logger_context=pass_logger_context,
                                                                                    pass_input_logger=pass_input_logger,
                                                                                    _matched=_matched)
            # create function context
            with self(context=_context,
                      init=True,
                      close=False,
                      defaults_config=defaults_config) as lc:
                # add function context to decorator_contexts
                self._decorator_contexts.append(lc)

                input_logger = ControlLogger.get_logger(f"{_name}.logger")

            def wrapper(*args, **kwargs):
                # perform time measurement
                start = None
                if func_config.measure_time:
                    start = time.perf_counter()

                # pass context and input logger to function
                if pass_logger_context is not None:
                    kwargs[_pass_logger_context] = lc
                if pass_input_logger is not None:
                    kwargs[_pass_input_logger] = input_logger

                # enter context
                with lc(close=False) as _lc:
                    result = self._log_function(*args,
                                                func_signature=func_signature,
                                                func_or_cls=func,
                                                self_or_cls=None,
                                                input_logger=input_logger,
                                                func_config=func_config,
                                                start=start,
                                                **kwargs)

                return result

            easy_logger.debug(f"Created context '{_context}' with name '{_name}' for function '{func.__name__}'")

            return wrapper

        return decorator

    def class_logger(self,
                     func_config: LogFuntionConfig | dict[str, any] = None,
                     func_configs: dict[str, LogFuntionConfig | dict[str, any]] = None,
                     name: str = None,
                     context: str = None,
                     cls_init_method: str = "__init__",
                     output_config: OutputLoggerConfig | dict[str, any] | Sequence[OutputLoggerConfig | dict[str, any]] = None,
                     input_config: InputLoggerConfig | dict[str, any] | Sequence[InputLoggerConfig | dict[str, any]] = None,
                     defaults_config: LoggerDefaultsConfig | dict[str, any] = None,
                     pass_logger_context: str | None = ...,
                     pass_input_logger: str | None = ...) -> Callable:

        if func_config is None:
            func_config = LogFuntionConfig()
        elif type(func_config) is dict:
            func_config = LogFuntionConfig(**func_config)
        elif not isinstance(func_config, LogFuntionConfig):
            raise TypeError(f"func_config must be of type {LogFuntionConfig.__name__} or dict.")

        if func_configs is None:
            func_configs = {}
        elif type(func_configs) is not dict:
            raise TypeError(f"func_configs must be of type dict[str, {LogFuntionConfig.__name__}] or dict[str, dict].")

        def decorator(cls) -> Callable:
            # check if function or type
            if not inspect.isclass(cls):
                raise NotImplementedError(f"Cant decorate function with {self.__class__.__name__}.class_logger(). Please use {self.__class__.__name__}.function_logger() instead.")

            # init matched
            _matched = []

            # set default values
            _name = name
            if name is None:
                _name = snake_case(cls.__name__)

            _context = context
            if context is None:
                _context = cls.__name__

            _pass_input_logger = pass_input_logger
            if pass_input_logger is Ellipsis:
                _pass_input_logger = "logger"

            _pass_logger_context = pass_logger_context
            if pass_logger_context is Ellipsis:
                _pass_logger_context = "logger_context"

            def logger_context_property(_self):
                for function_context in self.decorator_contexts:
                    if function_context.context == _context:
                        return function_context
                raise ValueError(f"Could not find context '{_context}' for class '{cls.__name__}'")

            def logger_property(_self):
                logger_context = ControlLogger.get_logger(f"{_name}.logger")
                return logger_context

            # check if function or type
            easy_logger.debug(f"Creating context '{_context}' with name '{_name}' for type '{cls.__name__}'")

            # create type context
            with self(context=_context,
                      init=True,
                      close=False,
                      defaults_config=defaults_config) as lc:
                # add type context to decorator_contexts
                self._decorator_contexts.append(lc)

            # set logger context property
            if _pass_logger_context is not None:
                if hasattr(cls, _pass_logger_context):
                    raise AttributeError(f"Class '{cls.__name__}' already has a 'logger_context' attribute.")
                setattr(cls, _pass_logger_context, property(logger_context_property))
            setattr(cls, "__class_logger_context__", property(logger_context_property))
            # set logger property
            if _pass_input_logger is not None:
                if hasattr(cls, _pass_input_logger):
                    raise AttributeError(f"Class '{cls.__name__}' already has a 'logger' attribute.")
                setattr(cls, _pass_input_logger, property(logger_property))
            setattr(cls, "__class_logger__", property(logger_property))

            # check cls_init_method in cls
            if not hasattr(cls, cls_init_method):
                raise ValueError(f"Could not find function '{cls_init_method}' in class '{cls.__name__}'.")
            __cls_init_method__ = getattr(cls, cls_init_method)
            if not inspect.isfunction(__cls_init_method__):
                raise ValueError(f"Could not find function '{cls_init_method}' in class '{cls.__name__}'.")

            # get signature
            cls_init_method_signature = inspect.signature(__cls_init_method__)

            # check if self or cls is in signature
            try:
                _matched, _self_or_cls = self._get_var_from_signature(signature=cls_init_method_signature,
                                                                      cls_type=None,
                                                                      prefer_name=["self", "cls"],
                                                                      _matched=_matched)
            except ValueError:
                _self_or_cls = None

            # overwrite cls_init_method
            def __class_logger_cls_init_method__(*args, **kwargs):
                # perform time measurement
                start = None
                if func_config.measure_time:
                    start = time.perf_counter()

                # enter context
                with lc(close=False) as _lc:
                    result = self._log_function(*args,
                                                func_signature=cls_init_method_signature,
                                                func_or_cls=__cls_init_method__,
                                                self_or_cls=_self_or_cls,
                                                input_logger=logger_property(None),
                                                func_config=func_config,
                                                start=start,
                                                **kwargs)

                return result

            setattr(cls, cls_init_method, __class_logger_cls_init_method__)

            # check if __del__ in cls
            # ToDo:

            # create logger methods for class
            for func_name in func_configs:
                # check if function in cls
                if not hasattr(cls, func_name):
                    raise ValueError(f"Could not find function '{func_name}' in class '{cls.__name__}'.")
                __func__ = getattr(cls, func_name)
                if hasattr(__func__, "__logger_class_method__"):
                    raise ValueError(f"Function '{func_name}' in class '{cls.__name__}' is already decorated with {self.__class__.__name__}.class_logger_method().")

                # decorate function
                __func__ = self.class_logger_method(func_config=func_configs[func_name])(__func__)

                # overwrite function
                setattr(cls, func_name, __func__)

            # get all logger_class_methods
            logger_class_methods = []
            for attr_name in dir(cls):
                # noinspection PyBroadException
                try:
                    attr = getattr(cls, attr_name)
                except Exception:
                    continue
                if hasattr(attr, "__logger_class_method__"):
                    logger_class_methods.append(attr)

            return cls

        return decorator

    def class_logger_method(self, func_config: LogFuntionConfig | dict[str, any] = None) -> Callable:

        if func_config is None:
            func_config = LogFuntionConfig()
        elif type(func_config) is dict:
            func_config = LogFuntionConfig(**func_config)
        elif not isinstance(func_config, LogFuntionConfig):
            raise TypeError(f"func_config must be of type {LogFuntionConfig.__name__} or dict.")

        def decorator(func: Callable | Type) -> Callable:
            # check if function or type
            if inspect.isclass(func):
                raise NotImplementedError(f"Cant decorate class with {self.__class__.__name__}.function_logger(). Please use {self.__class__.__name__}.class_logger() instead.")

            # init matched
            _matched = []

            if isinstance(func, classmethod):
                raise RuntimeError(f"Can't decorate class method with {self.__class__.__name__}.class_logger_method(). "
                                   f"Please use {self.__class__.__name__}.function_logger() instead.")
            if isinstance(func, staticmethod):
                raise RuntimeError(f"Can't decorate class method with {self.__class__.__name__}.class_logger_method(). "
                                   f"Please use {self.__class__.__name__}.function_logger() instead.")

            if isinstance(func, property):
                func = func.fget

            # get signature
            func_signature = inspect.signature(func)

            # check if self or cls is in signature
            try:
                _matched, _self_or_cls = self._get_var_from_signature(signature=func_signature,
                                                                      cls_type=None,
                                                                      prefer_name=["self", "cls"],
                                                                      _matched=_matched)
            except ValueError:
                _self_or_cls = None

            if _self_or_cls is None:
                raise ValueError(f"Could not find parameter 'self' or 'cls' in signature '{func_signature}'")

            def wrapper(self_or_cls, *args, **kwargs):
                # perform time measurement
                start = None
                if func_config.measure_time:
                    start = time.perf_counter()

                lc = self_or_cls.__class_logger_context__

                # add _self to kwargs
                args = [self_or_cls, *args]

                with lc(close=False) as _lc:
                    result = self._log_function(*args,
                                                func_signature=func_signature,
                                                func_or_cls=func,
                                                self_or_cls=_self_or_cls,
                                                input_logger=self_or_cls.__class_logger__,
                                                func_config=func_config,
                                                start=start,
                                                **kwargs)

                return result

            wrapper.__logger_class_method__ = True

            return wrapper

        return decorator


# set logger class
logging.setLoggerClass(InputLogger)



from __future__ import annotations

import dataclasses

from controllogger.enums.log_levels import LogLevels
from controllogger.misc.base_dataclass import BaseDataclass
from controllogger.misc.logger_defaults_config import LoggerDefaultsConfig


@dataclasses.dataclass
class BaseLoggerConfig(BaseDataclass):
    name: str  # Name of logger
    level: LogLevels | int | None = None  # Level of logger

    # log events
    log_events: bool | None = None  # Log events for logger. Is used to log events to logger. Default is False.
    log_init: bool | None = None  # Log init for input for logger. Is used to log the init event to loggers. Default is False.
    log_attach: bool | None = None  # Log attach for logger. Is used to log the attach event to logger. Default is False.
    log_detach: bool | None = None  # Log detach for logger. Is used to log the detach event to logger. Default is False.
    log_destroy: bool | None = None  # Log destroy for logger. Is used to log the destroy event to logger. Default is False.

    def __default_init__(self, logger_defaults_config: LoggerDefaultsConfig):
        # setup None values

        # level
        if self.level is None:
            self.level = logger_defaults_config.level

        # log_events
        if self.log_events is None:
            self.log_events = logger_defaults_config.log_events

        # log_init
        if self.log_init is None:
            if logger_defaults_config.log_init is None:
                self.log_init = self.log_events
            else:
                self.log_init = logger_defaults_config.log_init

        # log_attach
        if self.log_attach is None:
            if logger_defaults_config.log_attach is None:
                self.log_attach = self.log_events
            else:
                self.log_attach = logger_defaults_config.log_attach

        # log_detach
        if self.log_detach is None:
            if logger_defaults_config.log_detach is None:
                self.log_detach = self.log_events
            else:
                self.log_detach = logger_defaults_config.log_detach

        # log_destroy
        if self.log_destroy is None:
            if logger_defaults_config.log_destroy is None:
                self.log_destroy = self.log_events
            else:
                self.log_destroy = logger_defaults_config.log_destroy



from __future__ import annotations

import logging
import traceback
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum

from controllogger.enums.log_levels import LogLevels
from controllogger.enums.logger_event import LoggerEvent
from controllogger.logger.base_easy import BaseEasyLogger
from controllogger.misc.base_logger_config import BaseLoggerConfig
from controllogger.misc.context import LoggerContext
from controllogger.misc.easy_logger import easy_logger
from controllogger.misc.singleton import Singleton


# save if output logger on __new__ was called to prevent infinite loop
_on_new = False


class BaseLogger(BaseEasyLogger, ABC):
    _on_events: dict[str, list[callable]] = {}

    _ConfigClass = BaseLoggerConfig
    _my_type: str = "Logger"
    _control_logger_type_field: str = "None"

    __qwe__ = 123

    def __new__(cls, *args, **kwargs):
        global _on_new
        if not _on_new:
            _on_new = True
            control_logger = Singleton.get_by_type("ControlLogger")
            if cls._control_logger_type_field == "None":
                raise ValueError(f"Control logger type field not set for logger '{cls.__name__}'")
            logger_type = getattr(control_logger, cls._control_logger_type_field)
            __init__ = logger_type.__init__

            def __base_logger_init__(self, *a, **kw):
                if _on_new:
                    return
                self._use_control_logger_name = False
                __init__(self, *a, **kw)
                self.__post_init__()

            logger_type.__init__ = __base_logger_init__

            new_logger = logger_type(*args, **kwargs)
            _on_new = False
            return new_logger

        # register events
        cls.on_event(LoggerEvent.ATTACH)(cls.print_logger_attached)
        cls.on_event(LoggerEvent.INIT)(cls.print_logger_initialized)
        cls.on_event(LoggerEvent.DETACH)(cls.print_logger_detached)
        cls.on_event(LoggerEvent.DESTROY)(cls.print_logger_destroyed)

        return super().__new__(cls)

    def __init__(self, name: str = None, level: LogLevels | int = logging.NOTSET, config: _ConfigClass | dict[str, any] = None):
        # parse config
        if config is None:
            config = {}
        if type(config) is dict:
            if name is not None:
                config["name"] = name
            if level in [logging.NOTSET, None]:
                config["level"] = level

            config = self._ConfigClass(**config)
        elif type(config) is not self._ConfigClass:
            raise TypeError(f"config must be of type {self._ConfigClass.__name__} or dict[str, any]")

        self.config = config
        if self.control_logger.active_context is not None:
            self.config.__default_init__(logger_defaults_config=self.control_logger.active_context.logger_defaults_config)
        else:
            self.config.__default_init__(logger_defaults_config=self.control_logger.logger_defaults_config)

        if name is None:
            name = self.config.name
        if level not in [logging.NOTSET, None]:
            level = self.config.level

        # use control logger name
        if self._use_control_logger_name:
            name = self.control_logger.logger_defaults_config.name + "." + self.config.name
            if self.control_logger.active_context is not None:
                name = self.control_logger.active_context.logger_defaults_config.name + "." + self.config.name

        self._init = True
        if type(level) is Enum:
            level = level.value
        super().__init__(name, level)

        # check if logger already exists
        if self.control_logger.logger_exists(self.name):
            raise ValueError(f"Logger '{self.name}' already exists.")

        self._contexts: list[LoggerContext] = self.control_logger.active_contexts
        self._skip_handle = False  # resets after each log
        self._on_attaching = False

    def __post_init__(self):
        if self._init:
            # trigger init event
            self.trigger_event(LoggerEvent.INIT)
            self._init = False

    def __del__(self):
        # trigger destroy event
        self.trigger_event(LoggerEvent.DESTROY)

    @classmethod
    def _get_event_name(cls, event_name: str | LoggerEvent):
        if type(event_name) is not LoggerEvent:
            event_name = str(event_name)
            try:
                event_name = LoggerEvent[event_name.upper()]
            except KeyError:
                ...
        if event_name not in cls._on_events:
            cls._on_events[event_name] = []
        return event_name

    @classmethod
    def on_event(cls, event_name: LoggerEvent | str):

        event_name = cls._get_event_name(event_name)

        def decorator(func):
            easy_logger.debug(f"Registering event '{func.__name__}' for event '{event_name}' at logger '{cls.__name__}'")

            # get logger type name from calling class
            logger_type_str = func.__class__

            ...

            def wrapper(self, *args, **kwargs):
                control_logger = Singleton.get_by_type("ControlLogger")
                if not isinstance(self, cls):
                    return  # do nothing if self is not instance of cls
                easy_logger.debug(f"Calling method '{func.__name__}' for event '{event_name}' at logger '{cls.__name__}'")
                func(self, *args, **kwargs)
                easy_logger.debug(f"Called method '{func.__name__}' for event '{event_name}' at logger '{cls.__name__}'")

            cls._on_events[event_name].append(wrapper)

            return wrapper

        return decorator

    def trigger_event(self, event_name: LoggerEvent | str, *args, **kwargs):
        event_name = self._get_event_name(event_name)

        easy_logger.debug(f"Triggering event '{event_name}' for logger '{self.name}'", stacklevel=10)

        if event_name in self._on_events:
            for func in self._on_events[event_name]:
                func(self, *args, **kwargs)

        easy_logger.debug(f"Triggered event '{event_name}' for logger '{self.name}'", stacklevel=10)

    def print_logger_attached(self):
        easy_logger.debug(f"Attached logger '{self.name}' to control logger")
        if self.config.log_attach:
            self.print_header(
                name=f"{self._my_type} '{self.name}' attached at {datetime.now()}",
                level=LogLevels.DEBUG,
                lines=[str(self)],
            )

    def print_logger_initialized(self):
        easy_logger.debug(f"Initialized logger '{self.name}'")
        if self.config.log_init:
            self.print_header(
                name=f"{self._my_type} '{self.name}' initialized at {datetime.now()}",
                level=LogLevels.DEBUG,
                lines=[str(self)],
            )

    def print_logger_detached(self):
        easy_logger.debug(f"Detached logger '{self.name}' from control logger")
        if self.config.log_detach:
            self.print_header(
                name=f"{self._my_type} '{self.name}' detached at {datetime.now()}",
                level=LogLevels.DEBUG,
                lines=[str(self)],
            )

    def print_logger_destroyed(self):
        easy_logger.debug(f"Destroyed logger '{self.name}'")
        if self.config.log_destroy:
            self.print_header(
                name=f"{self._my_type} '{self.name}' destroyed at {datetime.now()}",
                level=LogLevels.DEBUG,
                lines=[str(self)],
            )

    @property
    @abstractmethod
    def attached(self) -> bool:
        ...

    @property
    def control_logger(self):
        """
        Returns the control logger singleton

        :return: ControlLogger
        """

        return Singleton.get_by_type("ControlLogger")

    @property
    def contexts(self) -> list[LoggerContext]:
        """
        Returns a copy of the list of context used by the logger

        :return: copy of the list of context used by the logger
        """

        return self._contexts.copy()

    @property
    def open(self) -> bool:
        if self.name not in self.control_logger.all_logger and not self._init:
            return False
        # if logger on creation
        return True

    @property
    def closed(self) -> bool:
        return not self.open

    @property
    def skip_handle(self) -> bool:
        return self._skip_handle

    @abstractmethod
    def attach(self) -> None:
        if len(self.contexts) == 0:
            easy_logger.debug(f"Attaching logger '{self.name}' to control logger")
        else:
            context_str = ", ".join([str(context) for context in self.contexts])
            easy_logger.debug(f"Attaching logger '{self.name}' to context '{context_str}'")
        self._on_attaching = True
        # trigger attach event
        self.trigger_event(LoggerEvent.ATTACH)
        self._on_attaching = False

    @abstractmethod
    def detach(self) -> None:
        if len(self.contexts) == 0:
            easy_logger.debug(f"Detaching logger '{self.name}' from control logger")
        else:
            context_str = ", ".join([str(context) for context in self.contexts])
            easy_logger.debug(f"Detaching logger '{self.name}' from context '{context_str}'")
        # trigger detach event
        self.trigger_event(LoggerEvent.DETACH)

    def handle(self, record):
        # raise error if logger is closed
        if not self.open:
            raise ValueError(f"Logger '{self.name}' is closed.")

        # raise error if logger is not attached
        if not self.attached and not self._on_attaching:
            raise ValueError(f"Logger '{self.name}' is not attached.")

        # skip handle if skip_handle is True
        skip_handle = self.skip_handle
        self._skip_handle = False
        if not skip_handle:
            return super().handle(record)

    def print_header(
            self,
            name: str,
            desc: str = None,
            lines: list[str] = None,
            level: LogLevels = LogLevels.DEBUG,
    ) -> None:
        """
        Prints a header to the terminal

        :param name: Name of the header
        :param desc: Description of the header
        :param lines: Lines to print
        :param level: Log level to use
        :return:
        """

        if not isinstance(level, int):
            if logging.raiseExceptions:
                raise TypeError("level must be an integer")
            else:
                return
        if self.isEnabledFor(level):
            self._log(level,
                      "[HEADER]",
                      (),
                      extra={"header": {"name": name,
                                        "desc": desc,
                                        "lines": lines,
                                        }
                             },
                      stack_info=False,
                      stacklevel=2)

    def print_traceback(self):
        """
        Prints the traceback to the logger
        :return:
        """
        tb = traceback.format_exc()
        self.exception(f"Traceback: \n{tb}")



from typing import NewType, Callable, Union

from controllogger.logger.input import InputLogger
from controllogger.misc.context import LoggerContext

LoggerContextImplementation = Union[NewType("LoggerContextImplementation", LoggerContext), Callable]
InputLoggerImplementation = NewType("InputLoggerImplementation", InputLogger)


class LoggerClass:
    logger_context: LoggerContextImplementation
    logger: InputLoggerImplementation



from enum import Enum


class Operator(str, Enum):
    EQUALS = "=="
    NOT_EQUALS = "!="
    CONTAINS = "contains"
    NOT_CONTAINS = "not_contains"
    STARTS_WITH = "starts_with"
    ENDS_WITH = "ends_with"
    GREATER_THAN = ">"
    GREATER_THAN_OR_EQUAL = ">="
    LESS_THAN = "<"
    LESS_THAN_OR_EQUAL = "<="
    IN = "in"
    NOT_IN = "not_in"



import sys
from pathlib import Path


def setup_exit(exit_code: int = 0, message: str = None):
    if message is not None:
        print(message)
    raise SystemExit(exit_code)


def get_version():
    with open(Path("src") / "controllogger" / "__version__", "r") as f:
        return f.read()


def change_version(new_version: str):
    with open(Path("src") / "controllogger" / "__version__", "w") as f:
        f.write(new_version)


def print_version():
    print(f"Current version: {get_version()}")
    setup_exit(0)


if __name__ == "__main__":
    old_version = None
    if "-bV" in sys.argv or "--build-version" in sys.argv:
        index = sys.argv.index("-bV" if "-bV" in sys.argv else "--build-version") + 1
        if index >= len(sys.argv):
            setup_exit(1, "Missing version number.")
        if sys.argv[index].startswith("-"):
            setup_exit(1, "Missing version number.")
        old_version = get_version()
        build_version = f"{old_version}.{sys.argv[index]}"
        change_version(build_version)
        sys.argv.pop(index)
        sys.argv.pop(index - 1)

    if "-v" in sys.argv or "--version" in sys.argv:
        print_version()

    print(f"Building version {get_version()}...")

    from setuptools import setup

    setup()
    print(f"Successfully built version {get_version()}.")
    if old_version is not None:
        change_version(old_version)



import dataclasses
import json
from pathlib import Path


@dataclasses.dataclass
class BaseDataclass:
    class JSONEncoder(json.JSONEncoder):
        def default(self, o):
            if isinstance(o, Path):
                return str(o)
            if dataclasses.is_dataclass(o):
                return dataclasses.asdict(o)
            if isinstance(o, object):
                return str(o)
            return super().default(o)

    def dict(self, exclude_none: bool = False) -> dict[str, any]:
        self_dict = dataclasses.asdict(self)

        # remove None values
        new_self_dict = {}
        if exclude_none:
            for key, value in self_dict.items():
                if value is None:
                    continue
                new_self_dict[key] = value
            self_dict = new_self_dict
        return self_dict

    def json(self, indent: int) -> str:
        self_dict = self.dict()

        self_json = json.dumps(self_dict, indent=indent, cls=self.JSONEncoder)

        return self_json



from enum import Enum


class LogLevels(int, Enum):
    """
    Available log levels for API
    """

    NOTSET = 0
    TRACE = 5
    DEBUG = 10
    INFO = 20
    WARNING = 30
    ERROR = 40
    CRITICAL = 50
